use crate::func::*;

struct CompositionClass<U, V>(U, V);

impl<U: WeakFunctor, V: WeakFunctor> WeakFunctor for CompositionClass<U, V> {
    type F<'a, A> = U::F<'a, V::F<'a, A>>;
}

impl<U: Functor, V: Functor> Functor for CompositionClass<U, V> {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        U::fmap(|ua| V::fmap(f, ua), fa)
    }

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        U::fmap(|ua| V::replace(ua, b), fa)
    }

    fn void<'a, A: 'a>(fa: Self::F<'a, A>) -> Self::F<'a, ()>
    where
        Self: 'a,
    {
        U::fmap(|ua| V::void(ua), fa)
    }
}

impl<U: ApplicativeLA2, V: ApplicativeSeq> ApplicativeSeq for CompositionClass<U, V> {
    fn seq<'a, A: 'a, B: 'a>(
        ff: Self::F<'a, impl 'a + FnOnce(A) -> B>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        U::la2(|uf, ua| V::seq(uf, ua), ff, fa)
    }
}

impl<U: ApplicativeLA2, V: ApplicativeLA2> ApplicativeLA2 for CompositionClass<U, V> {
    fn la2<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C>
    where
        Self: 'a,
    {
        U::la2(|ua, ub| V::la2(f, ua, ub), fa, fb)
    }
}

pub fn compose<'a>(f: impl 'a + Fn(i16) -> i32, g: impl 'a + Fn(i8) -> i16) -> impl 'a + Fn(i8) -> i32 {
    move |x| f(g(x))
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_compose() {
        assert_eq!(compose(|x| (x + 5).into(), |x| (x + 3).into())(2), 10);
    }
}
